# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from typing import Dict

import numpy as np
from pydefect.analyzer.band_edge_states import EdgeCharacters, EdgeCharacter
from pymatgen.core import Structure
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Procar, Vasprun, Outcar
from vise.analyzer.vasp.band_edge_properties import eigenvalues_from_vasprun


class MakeEdgeCharacters:
    def __init__(self,
                 procar: Procar,
                 vasprun: Vasprun,
                 outcar: Outcar,
                 neighboring_atom_indices):
        self.orbs = procar.data
        self.structure = vasprun.final_structure
        self.eigenvalues = eigenvalues_from_vasprun(vasprun)
        self.nelect = outcar.nelect
        self.mag = outcar.total_mag if outcar.total_mag else 0.0
        self.neighboring_atom_indices = neighboring_atom_indices

    @property
    def edge_characters(self):
        chars = []
        for spin, eigenvalues in self.eigenvalues.items():
            if spin == Spin.up:
                vbm_band_idx = int(round((self.nelect + self.mag) / 2)) - 1
            else:
                vbm_band_idx = int(round((self.nelect - self.mag) / 2)) - 1
            cbm_band_idx = vbm_band_idx + 1

            vbm = np.amax(eigenvalues[:, vbm_band_idx], axis=0)
            vbm_kpt_idx = np.argwhere(eigenvalues[:, vbm_band_idx] == vbm)[0][0]
            hob_bottom_e = np.amin(eigenvalues[:, vbm_band_idx], axis=0)

            cbm = np.amin(eigenvalues[:, cbm_band_idx], axis=0)
            cbm_kpt_idx = np.argwhere(eigenvalues[:, cbm_band_idx] == cbm)[0][0]
            lub_top_e = np.amax(eigenvalues[:, cbm_band_idx], axis=0)

            vbm_character = calc_orbital_character(self.orbs, self.structure, spin, vbm_kpt_idx, vbm_band_idx)
            cbm_character = calc_orbital_character(self.orbs, self.structure, spin, cbm_kpt_idx, cbm_band_idx)

            if self.neighboring_atom_indices:
                vbm_participation_ratio = calc_participation_ratio(self.orbs, spin, vbm_kpt_idx, vbm_band_idx, self.neighboring_atom_indices)
                cbm_participation_ratio = calc_participation_ratio(self.orbs, spin, cbm_kpt_idx, cbm_band_idx, self.neighboring_atom_indices)
            else:
                vbm_participation_ratio = None
                cbm_participation_ratio = None

            chars.append(EdgeCharacter(hob_bottom_e, lub_top_e, vbm, cbm, vbm_character, cbm_character, vbm_participation_ratio, cbm_participation_ratio))

        return EdgeCharacters(chars)


def calc_participation_ratio(orbitals: Dict[Spin, np.ndarray],
                             spin: Spin,
                             kpt_index: int,
                             band_index: int,
                             atom_indices: list) -> float:
    """ Returns sum of participation ratios at atom_indices sites

    The PROCAR data of the form below. It should VASP uses 1-based indexing,
    but all indices are converted to 0-based here.::
        { spin: np.ndarray accessed with (k-point index, band index,
                                          ion index, orbital index) }

    Note that the k-point weight is not considered, so all the k-points are
    treated equally.

    Return (float):
        float of the participation ratio.
    """
    sum_per_atom = np.sum(orbitals[spin][kpt_index, band_index, :, :], axis=1)
    return np.sum(sum_per_atom[atom_indices]) / np.sum(sum_per_atom)


def calc_orbital_character(orbitals,
                           structure: Structure,
                           spin: Spin,
                           kpt_index: int,
                           band_index: int):
    """ Consider the two pattern of orbitals.

    LORBIT 10 -> consider only "s", "p", "d" orbitals
    LORBIT >=11 -> consider "s", "px", "py", "pz",.. orbitals

    """

    def projection_sum(atom_indices: tuple, first: int, last: int):
        end = last + 1
        procar_sum = np.sum(orbitals[spin]
                            [kpt_index, band_index, atom_indices, first:end])
        return float(procar_sum)

    orbital_components = {}
    azimuthal = len(orbitals[Spin.up][0, 0, 0]) > 5

    for element in structure.symbol_set:
        # get list of index
        indices = structure.indices_from_symbol(element)
        if azimuthal:
            orbital_components[element] = \
                [round(projection_sum(indices, 0, 0), 3),
                 round(projection_sum(indices, 1, 3), 3),
                 round(projection_sum(indices, 4, 8), 3)]
            try:
                orbital_components[element].append(round(projection_sum(indices, 9, 16), 3))
            except KeyError:
                pass
        else:
            orbital_components[element] = \
                [round(projection_sum(indices, 0, 0), 3),
                 round(projection_sum(indices, 1, 1), 3),
                 round(projection_sum(indices, 2, 2), 3)]
    return orbital_components







# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pathlib import Path

from dash import Dash
from pydefect.analyzer.dash_components.cpd_energy_dash import CpdEnergyComponent
from pydefect.chem_pot_diag.chem_pot_diag import CpdPlotInfo
from monty.serialization import loadfn
import dash_html_components as html
import crystal_toolkit.components as ctc

app = Dash()

cwd = Path(__file__).parent

cpd = loadfn(str(cwd / "chem_pot_diag.json"))
info = CpdPlotInfo(cpd)
perfect = loadfn(str(cwd / "NaCl_perf_calc_results.json"))
defects = [loadfn(str(cwd / "NaCl_Va_Na1_0" / "calc_results.json")),
           loadfn(str(cwd / "NaCl_Va_Na1_-1" / "calc_results.json"))]
defect_entries = [loadfn(str(cwd / "NaCl_Va_Na1_0" / "defect_entry.json")),
                  loadfn(str(cwd / "NaCl_Va_Na1_-1" / "defect_entry.json"))]
corrections = [loadfn(str(cwd / "NaCl_Va_Na1_0" / "correction.json")),
               loadfn(str(cwd / "NaCl_Va_Na1_-1" / "correction.json"))]

component = CpdEnergyComponent(info, perfect, defects, defect_entries, corrections)

# example layout to demonstrate capabilities of component
my_layout = html.Div([component.layout])

ctc.register_crystal_toolkit(app=app, layout=my_layout, cache=None)

#app.run_server(mode='inline')


if __name__ == "__main__":
    app.run_server(debug=True, port=8086)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import string
from copy import deepcopy
from dataclasses import dataclass, InitVar
from itertools import chain
from pathlib import Path
from typing import Dict, Optional, Union, List, Tuple

import numpy as np
import pandas as pd
import yaml
from monty.json import MSONable
from monty.serialization import loadfn
from pydefect.error import PydefectError
from pydefect.util.error_classes import CpdNotSupportedError
from pymatgen.core import Composition, Element
from scipy.spatial.qhull import HalfspaceIntersection, QhullError
from tabulate import tabulate


alphabets = list(string.ascii_uppercase) + \
            ["A"+i for i in string.ascii_uppercase] + \
            ["B"+i for i in string.ascii_uppercase] + \
            ["C"+i for i in string.ascii_uppercase]


@dataclass
class CompositionEnergy(MSONable):
    composition: InitVar[Union[Composition, dict]]
    energy: float
    source: str

    def __post_init__(self, composition):
        if isinstance(composition, Composition):
            self.composition: Composition = composition
        else:
            self.composition: Composition = Composition.from_dict(composition)

    @property
    def abs_energy_per_atom(self):
        return self.energy / self.composition.num_atoms


@dataclass(repr=False)
class ChemPotDiag(MSONable):
    comp_energies: List[CompositionEnergy]
    target: InitVar[Union[Composition, dict]]
    vertex_elements: InitVar[Optional[List[Element]]] = None

    def __post_init__(self, target, vertex_elements):
        self.target: Composition = target if isinstance(target, Composition) \
            else Composition.from_dict(target)
        self.vertex_elements = vertex_elements or sorted(self.target.elements)

    def to_yaml(self, filename: str = "cpd.yaml") -> None:
        d = {"target": str(self.target).replace(" ", ""),
             "vertex_elements": [str(e) for e in self.vertex_elements]}
        for ce in self.comp_energies:
            key = str(ce.composition.iupac_formula).replace(" ", "")
            val = {"energy": ce.energy, "source": ce.source}
            d[key] = val
        Path(filename).write_text(yaml.dump(d))

    @classmethod
    def from_yaml(cls, filename: str = "cpd.yaml") -> "ChemPotDiag":
        d = loadfn(filename)
        target = d.pop("target")
        vertex_elements = [Element(e) for e in d.pop("vertex_elements")]
        composition_energies = []
        for k, v in d.items():
            composition_energies.append(
                CompositionEnergy(Composition(k), v["energy"], v["source"]))
        return cls(composition_energies, target, vertex_elements)

    @property
    def abs_energies_per_atom(self) -> Dict[Composition, float]:
        return {ce.composition.reduced_composition: ce.abs_energy_per_atom
                for ce in self.comp_energies}

    @property
    def all_compounds(self) -> List[Composition]:
        return list(self.abs_energies_per_atom.keys())

    @property
    def impurity_elements(self):
        result = set()
        for c in self.all_compounds:
            result.update(set(c.elements))
        result.difference_update(set(self.vertex_elements))
        return list(result)

    @property
    def dim(self):
        return len(self.vertex_elements)

    @property
    def vertex_elements_abs_energies_per_atom(self):
        result = {}
        for k, v in self.abs_energies_per_atom.items():
            if set(k.elements).issubset(set(self.vertex_elements)):
                result[k] = v
        return result

    def impurity_abs_energy_per_atom(self, elem: Element):
        for k, v in self.abs_energies_per_atom.items():
            if len(k.elements) == 1 and k.elements[0] == elem:
                return v
        raise ValueError(f"{elem} is not included.")

    @property
    def offset_to_abs(self):
        result = []
        for vertex_element in self.vertex_elements:
            target = Composition({vertex_element: 1.0}).reduced_composition
            candidates = \
                filter(lambda x: x[0] == target,
                       self.vertex_elements_abs_energies_per_atom.items())
            try:
                result.append(min([x[1] for x in candidates]))
            except ValueError:
                raise NoElementEnergyError
        return result

    @property
    def rel_energies(self):
        result = {}
        for c, e in self.vertex_elements_abs_energies_per_atom.items():
            sub = sum(f * offset for f, offset
                      in zip(self.atomic_fractions(c), self.offset_to_abs))
            result[c] = e - sub

        return result

    def atomic_fractions(self, c):
        return [c.fractional_composition[e] for e in self.vertex_elements]

    @property
    def vertex_coords(self):
        if self.dim == 1:
            return [np.array([0.0])]

        large_minus_number = -1e5
        hs = self.get_half_space_intersection(large_minus_number)

        result = []
        for intersection in hs.intersections:
            if min(intersection) != large_minus_number:
                result.append(intersection)

        return result

    @property
    def lowest_relative_energy(self):
        return min(chain(*self.vertex_coords))

    def get_half_space_intersection(self, min_range, feasible_dist=1.0):
        half_spaces = []
        for c, e in self.rel_energies.items():
            half_spaces.append(self.atomic_fractions(c) + [-e])
        for i in range(self.dim):
            x = [0.0] * self.dim
            x[i] = -1.0
            x.append(min_range)
            half_spaces.append(x)
        # When used in vertex_coords  min_range is large_minum_number
        feasible_point = np.array([min_range + feasible_dist] * self.dim,
                                  dtype=float)
        hs = HalfspaceIntersection(np.array(half_spaces), feasible_point)

        return hs

    @property
    def target_vertices(self) -> Dict[str, List[float]]:
        label = iter(alphabets)
        fractions = self.atomic_fractions(self.target)
        energy = self.rel_energies[self.target]
        return {next(label): c for c in self.vertex_coords
                if on_composition(fractions, c, energy)}

    @property
    def target_vertex_list_dataframe(self):
        index = []
        result = []
        for k, v in self.target_vertices.items():
            index.append(k)
            result.append([])
            for k2, v2 in zip(self.vertex_elements, v):
                result[-1].append(round(v2, 3))

            for ie in self.impurity_elements:
                competing_comp_for_impurity, energy = self.impurity_rel_energy(ie, k)
                result[-1].append(round(energy, 3))
                comp_name = \
                    competing_comp_for_impurity.composition.reduced_formula
                result[-1].append(comp_name)

        columns = [f"mu_{e}" for e in self.vertex_elements]
        for e in self.impurity_elements:
            columns += [f"mu_{e}", f"{e} competing phase"]
        return pd.DataFrame(result, index=index, columns=columns)

    def __repr__(self):
        return tabulate(self.target_vertex_list_dataframe,
                        headers='keys', tablefmt='psql')

    def target_abs_chem_pot_dict(self, label) -> Dict[Element, float]:
        rel_chem_pots = self.target_vertices[label]
        abs_chem_pots = [x + y for x, y in zip(rel_chem_pots,
                                               self.offset_to_abs)]
        return dict(zip(self.vertex_elements, abs_chem_pots))

    def abs_chem_pot_dict(self, label) -> Dict[Element, float]:
        result = deepcopy(self.target_abs_chem_pot_dict(label))
        for e in self.impurity_elements:
            _, result[e] = self.impurity_abs_energy(e, label)
        return result

    def label_at_rich_condition(self, element: Element) -> str:
        """Return abs_chem_pot_dict at condition where an element is rich."""
        ref_e = -float("inf")
        target_index = self.vertex_elements.index(element)
        result = None
        for label, vertex in self.target_vertices.items():
            if vertex[target_index] > ref_e:
                result, ref_e = label, vertex[target_index]
        return result

    def label_at_poor_condition(self, element: Element) -> str:
        """Return abs_chem_pot_dict at condition where an element is poor."""
        ref_e = float("inf")
        target_index = self.vertex_elements.index(element)
        result = None
        for label, vertex in self.target_vertices.items():
            if vertex[target_index] < ref_e:
                result, ref_e = label, vertex[target_index]
        return result

    def impurity_abs_energy(self, element: Element, label: str):
        if element not in self.impurity_elements:
            raise ValueError(
                f"Element {element} not in impurities. List of impurities are "
                f"{self.impurity_elements}.")
        comp_set = set(self.vertex_elements) | {element}
        print(comp_set)
        competing_comp_e = None
        y = float("inf")
        for ce in self.comp_energies:
            if element in set(ce.composition.elements):
                if set(ce.composition.elements).issubset(comp_set) is False:
                    raise CpdNotSupportedError(
                        "Other element(s) than host elements exists.")
                abs_chem_pot = self.target_abs_chem_pot_dict(label)
                mu = ce.energy
                comp_d = ce.composition.as_dict()
                for ve in self.vertex_elements:
                    mu -= comp_d[str(ve)] * abs_chem_pot[ve]
                mu /= comp_d[str(element)]
                if mu < y:
                    competing_comp_e, y = ce, mu
        if competing_comp_e is None:
            raise CpdNotSupportedError(f"No compounds with element {element}.")
        return competing_comp_e, y

    def impurity_rel_energy(self, element: Element, label: str):
        competing_comp_e, y = self.impurity_abs_energy(element, label)
        return competing_comp_e,  y - self.impurity_abs_energy_per_atom(element)

    @property
    def target_formation_energy(self):
        return self.rel_energies[self.target]


class CpdPlotInfo:
    def __init__(self,
                 cpd: ChemPotDiag,
                 min_range: Optional[float] = None):
        self.cpd = cpd
        self.min_range = min_range or self.cpd.lowest_relative_energy * 1.1

        self.dim = cpd.dim
        self.comp_vertices = self._get_comp_vertices(self.min_range)

    def _get_comp_vertices(self, min_range
                           ) -> Dict[Composition, List[List[float]]]:
        try:
            hs = self.cpd.get_half_space_intersection(min_range)
        except QhullError:
            hs = self.cpd.get_half_space_intersection(min_range,
                                                      feasible_dist=0.01)

        intersections = hs.intersections.tolist()
        result = {}
        for c, e in self.cpd.rel_energies.items():
            def on_comp(coord):
                return on_composition(self.cpd.atomic_fractions(c), coord, e)

            coords = [[round(j, ndigits=5) for j in i]
                      for i in intersections if on_comp(i)]

            if coords:
                result[c] = coords

        return result

    @property
    def comp_centers(self):
        return {c: np.average(np.array(v), axis=0).tolist()
                for c, v in self.comp_vertices.items()}

    def atomic_fractions(self, composition: Composition):
        return [composition.get_atomic_fraction(e)
                for e in self.cpd.vertex_elements]


def on_composition(atomic_fractions, coord, energy) -> bool:
    diff = sum([x * y for x, y in zip(atomic_fractions, coord)]) - energy
    return abs(diff) < 1e-8


def replace_comp_energy(chem_pot_diag: ChemPotDiag,
                        replaced_comp_energies: List[CompositionEnergy]):
    new_comp_energies = []
    for ce in chem_pot_diag.comp_energies:
        for replaced_comp_energy in replaced_comp_energies:
            if (ce.composition.reduced_composition
                    == replaced_comp_energy.composition.reduced_composition):
                new_comp_energies.append(replaced_comp_energy)
                break
        else:
            new_comp_energies.append(ce)
    chem_pot_diag.comp_energies = new_comp_energies


class NoElementEnergyError(PydefectError):
    pass



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path
from shutil import copyfile
from typing import List

import yaml
from pydefect.cli.vasp.molecules.molecules import MOLECULE_DATA
from pymatgen.core import Composition

mol_dir = Path(__file__).parent / "molecules"


def make_poscars_from_query(materials_query: List[dict], path: Path) -> None:

    for m in materials_query:
        reduced_formula = Composition(m["full_formula"]).reduced_formula

        if reduced_formula in MOLECULE_DATA.keys():
            dirname = path / f"mol_{reduced_formula}"
            if not dirname.exists():
                dirname.mkdir()
                copyfile(mol_dir / reduced_formula / "POSCAR",
                         dirname / "POSCAR")
                copyfile(mol_dir / reduced_formula / "prior_info.yaml",
                         dirname / "prior_info.yaml")

        else:
            m_path = path / f"{reduced_formula}_{m['task_id']}"
            m_path.mkdir()
            m["structure"].to(filename=m_path / "POSCAR")

            d = {"total_magnetization": m["total_magnetization"],
                 "band_gap": m["band_gap"],
                 "data_source": m["task_id"]}

            prior_info = m_path / "prior_info.yaml"
            prior_info.write_text(yaml.dump(d), None)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
from pydefect.analyzer.band_edge_states import BandEdgeEigenvalues
from pydefect.defaults import defaults
from pymatgen.io.vasp import Vasprun


def make_band_edge_eigenvalues(vasprun: Vasprun, vbm: float, cbm: float):
    """

    eigenvalues[Spin][k-index][band-index] = [energy, occupation]
    """
    eigval_range = defaults.eigval_range
    kpt_coords = [tuple(coord) for coord in vasprun.actual_kpoints]

    max_energy_by_spin = []
    min_energy_by_spin = []

    for e in vasprun.eigenvalues.values():
        max_energy_by_spin.append(np.amax(e[:, :, 0], axis=0))
        min_energy_by_spin.append(np.amin(e[:, :, 0], axis=0))

    max_energy_by_band = np.amax(np.vstack(max_energy_by_spin), axis=0)
    min_energy_by_band = np.amin(np.vstack(min_energy_by_spin), axis=0)

    lower_idx = np.argwhere(max_energy_by_band > vbm - eigval_range)[0][0]
    upper_idx = np.argwhere(min_energy_by_band < cbm + eigval_range)[-1][-1]

    e_and_occu = []
    for e in vasprun.eigenvalues.values():
        e_and_occu.append(e[:, lower_idx:upper_idx + 1, :].tolist())

    return BandEdgeEigenvalues(energies_and_occupations=e_and_occu,
                               kpt_coords=kpt_coords,
                               # need to convert numpy.int64 to int for mongoDB.
                               lowest_band_index=int(lower_idx))



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pydefect.analyzer.calc_results import CalcResults
from pydefect.defaults import defaults
from pymatgen.io.vasp import Vasprun, Outcar
from vise.analyzer.vasp.band_edge_properties import VaspBandEdgeProperties
from vise.util.structure_symmetrizer import StructureSymmetrizer


def make_calc_results_from_vasp(vasprun: Vasprun,
                                outcar: Outcar) -> CalcResults:
    structure = vasprun.final_structure
    magnetization = outcar.total_mag or 0.0

    symmetrizer = StructureSymmetrizer(
        structure,
        symprec=defaults.symmetry_length_tolerance,
        angle_tolerance=defaults.symmetry_angle_tolerance,
        time_reversal=abs(magnetization) > defaults.integer_criterion)
    band_edge_prop = VaspBandEdgeProperties(
        vasprun, outcar, defaults.integer_criterion)

    return CalcResults(structure=structure,
                       site_symmetry=symmetrizer.point_group,
                       energy=outcar.final_energy,
                       magnetization=magnetization,
                       kpoint_coords=vasprun.actual_kpoints,
                       kpoint_weights=vasprun.actual_kpoints_weights,
                       potentials=[-p for p in outcar.electrostatic_potential],
                       vbm_info=band_edge_prop.vbm_info,
                       cbm_info=band_edge_prop.cbm_info,
                       fermi_level=vasprun.efermi,
                       electronic_conv=vasprun.converged_electronic,
                       ionic_conv=vasprun.converged_ionic)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List

import numpy as np
from monty.json import MSONable
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class Unitcell(MSONable, ToJsonFileMixIn):
    vbm: float
    cbm: float
    ele_dielectric_const: List[List[float]]
    ion_dielectric_const: List[List[float]]

    @property
    def dielectric_constant(self):
        total = (np.array(self.ele_dielectric_const)
                 + np.array(self.ion_dielectric_const))
        return total.tolist()




#!/usr/bin/env python
# coding: utf-8

# In[1]:


from crystal_toolkit.helpers.layouts import Columns, Column
from crystal_toolkit.settings import SETTINGS
from jupyter_dash import JupyterDash
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.dash_components.cpd_energy_dash import CpdEnergyComponent
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag, CpdPlotInfo,     CompositionEnergy
from pydefect.corrections.manual_correction import ManualCorrection
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import Composition, Structure, Lattice, Element
import dash_html_components as html
import crystal_toolkit.components as ctc
from dash.dependencies import Input, Output, State
import json

app = JupyterDash(suppress_callback_exceptions=True,
               assets_folder=SETTINGS.ASSETS_PATH)
from vise.analyzer.band_edge_properties import BandEdge

comp_energies = [
    CompositionEnergy(Composition("Mg"), 0.0, "a"),
    CompositionEnergy(Composition("Ca"), 0.0, "a"),
    CompositionEnergy(Composition("Sr"), 0.0, "a"),
    CompositionEnergy(Composition("O"), 0.0, "a"),
    CompositionEnergy(Composition("H"), 0.0, "a"),
#    CompositionEnergy(Composition("MgCaO3"), -100.0, "a"),
    CompositionEnergy(Composition("MgCaSrO3"), -100.0, "a"),
]
#cpd = ChemPotDiag(comp_energies, target=Composition("MgCaO3"))
cpd = ChemPotDiag(comp_energies, target=Composition("MgCaSrO3"))
cpd_plot_info = CpdPlotInfo(cpd)


# In[2]:


print(cpd.target.elements)
print(cpd.dim)
print(cpd.target_vertices)
print(cpd.all_compounds)
print(cpd.impurity_abs_energy(Element.H, label="A"))


# In[3]:


structure = Structure(Lattice.cubic(1), species=["O"] * 2, coords=[[0]*3]*2)
defect_structure = Structure(Lattice.cubic(1), species=["O"] * 1, coords=[[0]*3])

common = dict(site_symmetry="1",
             magnetization=0.0,
             kpoint_coords=[[0]*3],
             kpoint_weights=[1.0],
             potentials=[0.0],
             vbm_info=BandEdge(0.0),
             cbm_info=BandEdge(1.0),
             fermi_level=0.0)

perfect = CalcResults(structure=structure,energy=0, **common)
defects = [CalcResults(structure=defect_structure, energy=1.0, **common),
          CalcResults(structure=defect_structure, energy=0.5, **common)]

de_common = dict(name="Va_O1",
                structure=defect_structure, site_symmetry="1",
                perturbed_structure=defect_structure, defect_center=[[0]*3])

defect_entries = [DefectEntry(charge=0, **de_common),
                 DefectEntry(charge=1, **de_common)]

corrections = [ManualCorrection(correction_energy=1.0),
              ManualCorrection(correction_energy=1.0)]

cpd_e_component = CpdEnergyComponent(cpd_plot_info,
                                     perfect,
                                     defects,
                                     defect_entries,
                                     corrections,
                                     unitcell_vbm=0.0,
                                     unitcell_cbm=1.0)

my_layout = html.Div([Column(cpd_e_component.layout)])
ctc.register_crystal_toolkit(app=app, layout=my_layout, cache=None)


# In[5]:


app.run_server(port=8099)
#app.run_server(mode='inline', port=8096)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections.abc import Set
from typing import Iterator, List, Set as typeSet

from monty.serialization import loadfn, dumpfn
from pydefect.input_maker.defect import screen_simple_defect, SimpleDefect


class DefectSet(Set):
    def __init__(self, defects: typeSet[SimpleDefect]):
        self.defects = defects

    def __contains__(self, defect: object) -> bool:
        return defect in self.defects

    def __len__(self) -> int:
        return len(self.defects)

    def __iter__(self) -> Iterator:
        yield from self.defects

    def to_yaml(self, filename: str = "defect_in.yaml") -> None:
        d = {defect.name: list(defect.charges) for defect in self.defects}
        dumpfn(d, filename)

    @classmethod
    def from_yaml(cls, filename: str = "defect_in.yaml") -> "DefectSet":
        d = loadfn(filename)
        names = set()
        for name, charges in d.items():
            in_name, out_name = name.split("_")
            names.add(SimpleDefect(in_name, out_name, tuple(charges)))
        return cls(names)


def screen_defect_set(defect_set: DefectSet, keywords: List[str]):
    result = set()
    for defect in defect_set:
        screened = screen_simple_defect(defect, keywords)
        if screened:
            result.add(screened)
    return DefectSet(result)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from pathlib import Path
from typing import Tuple, Optional

import numpy as np
import yaml
from monty.json import MSONable
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator
from pymatgen.core import IStructure
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.structure_symmetrizer import StructureSymmetrizer


@dataclass(frozen=True)
class DefectEntry(MSONable, ToJsonFileMixIn):
    name: str
    charge: int
    structure: IStructure
    perturbed_structure: Optional[IStructure]
    site_symmetry: str
    defect_center: Tuple[float, float, float]

    @property
    def anchor_atom_index(self) -> int:
        """ Returns an index of atom that is the farthest from the defect.

         Only the first occurrence is returned when using argmax.
         docs.scipy.org/doc/numpy-1.15.0/reference/generated/numpy.argmax.html
         """
        lattice = self.structure.lattice
        all_coords = self.structure.frac_coords
        dist_set = lattice.get_all_distances(self.defect_center, all_coords)[0]
        return int(np.argmax(dist_set))

    @property
    def anchor_atom_coords(self) -> np.ndarray:
        return self.structure[self.anchor_atom_index].frac_coords

    @property
    def perturbed_site_indices(self):
        result = []
        for i, (s1, s2) in enumerate(zip(self.structure,
                                         self.perturbed_structure)):
            if (s1.frac_coords != s2.frac_coords).any():
                result.append(i)
        return result

    @property
    def full_name(self):
        return "_".join([self.name, str(self.charge)])

    def to_prior_info(self, filename):
        d = {"charge": self.charge}
        Path(filename).write_text(yaml.dump(d))


def make_defect_entry(name: str,
                      charge: int,
                      perfect_structure: IStructure,
                      defect_structure: IStructure):

    analyzer = DefectStructureComparator(perfect_structure, defect_structure)

    species = []
    frac_coords = []
    for d, p in enumerate(analyzer.p_to_d):
        if p is None:
            site = defect_structure[d]
        else:
            site = perfect_structure[p]
        species.append(site.specie)
        frac_coords.append(site.frac_coords)

    initial_structure = IStructure(perfect_structure.lattice,
                                   species, frac_coords)
    symmetrizer = StructureSymmetrizer(initial_structure)

    return DefectEntry(name, charge, initial_structure, None,
                       site_symmetry=symmetrizer.point_group,
                       defect_center=tuple(analyzer.defect_center_coord))



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

import pytest
from monty.serialization import loadfn
from pydefect.analyzer.band_edge_states import BandEdgeEigenvalues
from pydefect.analyzer.eigenvalue_plotter import EigenvalueMplPlotter, \
    EigenvaluePlotlyPlotter
from vise.util.dash_helper import show_png

try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


eig = BandEdgeEigenvalues(energies_and_occupations=[
    [[[0.0, 1.0], [0.5, 0.5], [1.0, 0.0]], [[0.0, 1.0], [0.5, 0.5], [1.0, 0.0]]],
    [[[0.0, 1.0], [0.5, 0.5], [1.0, 0.0]], [[0.0, 1.0], [0.5, 0.5], [1.0, 0.0]]],
],
    kpt_coords=[(0.0, 0.0, 0.0), (0.25, 0.0, 0.0)],
    lowest_band_index=10)


@pytest.fixture
def eigenvalue_mpl_plotter():
    return EigenvalueMplPlotter(title="test", band_edge_eigenvalues=eig,
                                supercell_vbm=0.1, supercell_cbm=0.9)


def test_plot(eigenvalue_mpl_plotter):
    eigenvalue_mpl_plotter.construct_plot()
    eigenvalue_mpl_plotter.plt.show()


def test_plot_with_actual_file():
    eig = loadfn(Path(__file__).parent / "band_edge_eigenvalues.json")
    plotter = EigenvalueMplPlotter(title="test", band_edge_eigenvalues=eig,
                                   supercell_vbm=0.5, supercell_cbm=2.6)
    plotter.construct_plot()
    plotter.plt.show()


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="psutil does not exist")
def test_plotly_with_actual_file():
    eig = loadfn(Path(__file__).parent / "band_edge_eigenvalues.json")
    plotter = EigenvaluePlotlyPlotter(title="test", band_edge_eigenvalues=eig,
                                      supercell_vbm=0.5, supercell_cbm=2.6)
    fig = plotter.create_figure()
    # fig.show()
    show_png(fig)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import argparse
import sys
from itertools import groupby
from pathlib import Path
from typing import List, Optional, Union

from monty.serialization import loadfn
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag, CompositionEnergy, \
    CpdPlotInfo
from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiagPlotly3DMplPlotter, \
    ChemPotDiagPlotly2DMplPlotter
from pydefect.util.mp_tools import MpQuery
from pymatgen.core import Composition, Element
from vise.atom_energies.atom_energy import mp_energies
from vise.util.logger import get_logger

parent = Path(__file__).parent
logger = get_logger(__name__)


def make_chem_pot_diag_from_mp(target: Union[Composition, str],
                               additional_elements: List[str] = None,
                               vertex_elements: List[str] = None,
                               atom_energy_yaml: Optional[str] = None):
    """Obtain the energies from Materials Project.

    When the atom_energy_yaml is provided, the total energies are aligned
    via atom energies.

    """
    properties = ["task_id", "full_formula", "final_energy"]
    target = target if isinstance(target, Composition) else Composition(target)
    elements = target.chemical_system.split("-")
    vertex_elements = vertex_elements or elements
    vertex_elements = [Element(e) for e in vertex_elements]
    if additional_elements:
        elements.extend(additional_elements)
    query = MpQuery(elements, properties=properties)
    comp_es = []
    if atom_energy_yaml:
        energies = loadfn(atom_energy_yaml)
        diff = {e: energies[e] - mp_energies[e] for e in elements}
    else:
        diff = None

    for m in query.materials:
        energy = m["final_energy"]
        if diff:
            for k, v in Composition(m["full_formula"]).as_dict().items():
                energy += diff[k] * v
        comp_es.append(CompositionEnergy(
            Composition(m["full_formula"]), energy, m["task_id"]))

    comp_es = remove_higher_energy_comp(comp_es)

    return ChemPotDiag(comp_es, target, vertex_elements)


def remove_higher_energy_comp(comp_energies: List[CompositionEnergy]):
    result = []
    for _, grouped_comp_energies in groupby(
            comp_energies, key=lambda x: x.composition.reduced_formula):
        result.append(min(list(grouped_comp_energies),
                          key=lambda y: y.abs_energy_per_atom))
    return result


def parse_args(args):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-e", "--elements", type=str, nargs="+")
    parser.add_argument("-he", "--host_elements", type=Element, nargs="+")
    parser.add_argument("-t", "--target", type=Composition)
    parser.add_argument("-f", "--functional", type=str)
    return parser.parse_args(args)


if __name__ == "__main__":
    x = parse_args(sys.argv[1:])
    cpd = make_chem_pot_diag_from_mp(x.elements, x.target,
                                     x.host_elements, x.functional)
    cpd.to_yaml()
    if cpd.dim == 2:
        plotter = ChemPotDiagPlotly2DMplPlotter(CpdPlotInfo(cpd))
    elif cpd.dim == 3:
        plotter = ChemPotDiagPlotly3DMplPlotter(CpdPlotInfo(cpd))
    plotter.figure.show()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from dataclasses import dataclass
from typing import List, Optional

from monty.json import MSONable
from pymatgen.core import IStructure
from vise.analyzer.band_edge_properties import BandEdge
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class CalcResults(MSONable, ToJsonFileMixIn):
    structure: IStructure
    site_symmetry: str
    energy: float
    magnetization: float
    kpoint_coords: List[List[float]]
    kpoint_weights: List[float]
    # potential acting on the positive unit charge, whose sign is reserved from
    # vasp convention of atomic site potential.
    potentials: List[float]
    vbm_info: BandEdge
    cbm_info: BandEdge
    fermi_level: float
    electronic_conv: Optional[bool] = None
    ionic_conv: Optional[bool] = None

    @property
    def vbm(self):
        return self.vbm_info.energy

    @property
    def cbm(self):
        return self.cbm_info.energy





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.


import argparse
import sys
import warnings
from pathlib import Path

from monty.serialization import loadfn
from pydefect import __version__
from pydefect.cli.main_tools import str_int_to_int
from pydefect.cli.vasp.main_function import make_supercell, make_defect_set, \
    make_defect_entries, make_unitcell, make_competing_phase_dirs, \
    make_chem_pot_diag, make_efnv_correction_from_vasp, print_file, \
    make_calc_results, make_defect_formation_energy, make_defect_eigenvalues, \
    make_edge_characters, make_edge_states, \
    append_interstitial_to_supercell_info, pop_interstitial_from_supercell_info, \
    plot_chem_pot_diag, make_gkfo_correction_from_vasp, calc_defect_structure_info
from pydefect.defaults import defaults
from pymatgen.core import IStructure, Composition, Structure
from pymatgen.io.vasp import Vasprun, Outcar
from pymatgen.io.vasp.inputs import UnknownPotcarWarning

warnings.simplefilter('ignore', UnknownPotcarWarning)


def parse_args(args):

    parser = argparse.ArgumentParser(
        description="""                            
    pydefect is a package that helps researchers to do first-principles point 
    defect calculations with the VASP code.""",
        epilog=f"Author: Yu Kumagai Version: {__version__}")

    subparsers = parser.add_subparsers()

    # ++ parent parser: dirs
    dirs_parser = argparse.ArgumentParser(
        description="", add_help=False)
    dirs_parser.add_argument(
        "-d", "--dirs", nargs="+", type=Path,
        help="Directory paths to be parsed.")

    # ++ parent parser: pcr
    si_parser = argparse.ArgumentParser(
        description="", add_help=False)
    si_parser.add_argument(
        "-s", "--supercell_info", required=True, type=loadfn,
        help="Path to the supercell_info.json file.")

    # ++ parent parser: pcr
    pcr_parser = argparse.ArgumentParser(
        description="", add_help=False)
    pcr_parser.add_argument(
        "-pcr", "--perfect_calc_results", required=True, type=loadfn,
        help="Path to the calc_results.json for the perfect supercell.")

    # ++ parent parser: unitcell
    unitcell_parser = argparse.ArgumentParser(
        description="", add_help=False)
    unitcell_parser.add_argument(
        "-u", "--unitcell", required=True, type=loadfn,
        help="Path to the unitcell.json file.")

    # -- print ------------------------------------------------
    parser_print = subparsers.add_parser(
        name="print",
        description="Print the json file info.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['p'])

    parser_print.add_argument(
        "-f", "--filename", dest="obj", required=True, type=loadfn,
        help="Path to the json file to be printed.")

    parser_print.set_defaults(func=print_file)

    # -- unitcell ------------------------------------------------
    parser_unitcell = subparsers.add_parser(
        name="unitcell",
        description="Parse unitcell info and create unitcell.json",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['u'])

    parser_unitcell.add_argument(
        "-vb", "--vasprun_band", required=True, type=Vasprun,
        help="vasprun.xml file of band structure calculation.")
    parser_unitcell.add_argument(
        "-ob", "--outcar_band", required=True, type=Outcar,
        help="OUTCAR file of band structure calculation.")
    parser_unitcell.add_argument(
        "-odc", "--outcar_dielectric_clamped", required=True, type=Outcar,
        help="OUTCAR file for ion-clamped dielectric constant.")
    parser_unitcell.add_argument(
        "-odi", "--outcar_dielectric_ionic", required=True, type=Outcar,
        help="OUTCAR file for dielectric constant of ionic contribution.")

    parser_unitcell.set_defaults(func=make_unitcell)

    # -- make_poscars ------------------------------------------------
    parser_make_poscars = subparsers.add_parser(
        name="make_poscars",
        description="Make poscar files in directories retrieved from "
                    "the Materials Project.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['mp'])

    parser_make_poscars.add_argument(
        "-e", "--elements", required=True, nargs="+", type=str,
        help="Element names considered in chemical potential diagram. ")
    parser_make_poscars.add_argument(
        "--e_above_hull", default=defaults.e_above_hull, type=float,
        help="Energy above hull in eV/atom.")

    parser_make_poscars.set_defaults(func=make_competing_phase_dirs)

    # -- make_cpd ------------------------------------------------
    parser_mcpd = subparsers.add_parser(
        name="make_cpd",
        parents=[dirs_parser],
        description="Make chemical potential diagram.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['mcpd'])

    parser_mcpd.add_argument(
        "-y", "--yaml", type=str, default="cpd.yaml",
        help="Saved yaml file name.")
    parser_mcpd.add_argument(
        "-t", "--target", type=Composition, required=True,
        help="Target composition, e.g., MgO.")
    parser_mcpd.add_argument(
        "-e", "--elements", type=str, nargs="+",
        help="Element names considered in chemical potential diagram. Used for"
             "creating the diagram from mp data with atom energy alignment.")
    parser_mcpd.add_argument(
        "-a", "--atom_energy_yaml", type=str,
        help="Yaml file storing atom energies for energy alignment.")
    parser_mcpd.add_argument(
        "-u", "--update", action="store_true",
        help="Used when one wants to update the cpd.yaml with calculation "
             "results.")
    parser_mcpd.add_argument(
        "-v", "--vertices", nargs="+",
        help="Set vertex element names for chemical potential diagram, e.g., "
             "Mg Al O.")

    parser_mcpd.set_defaults(func=make_chem_pot_diag)

    # -- plot_cpd ------------------------------------------------
    parser_pcpd = subparsers.add_parser(
        name="plot_cpd",
        description="Plot chemical potential diagram and make pdf file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pcpd'])

    parser_pcpd.add_argument(
        "-y", "--yaml", default="cpd.yaml", type=str,
        help="cpd.yaml file name")

    parser_pcpd.set_defaults(func=plot_chem_pot_diag)

    # -- supercell ------------------------------------------------
    parser_supercell = subparsers.add_parser(
        name="supercell",
        description="Make SPOSCAR file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['s'])

    parser_supercell.add_argument(
        "-p", "--unitcell", default="POSCAR", type=IStructure.from_file,
        help="Base POSCAR, which must be the standardized primitive cell.")
    parser_supercell.add_argument(
        "--matrix", nargs="+", type=int,
        help="Supercell matrix applied to the conventional cell. "
             "1, 3 or 9 components are accepted.")
    parser_supercell.add_argument(
        "--min_atoms", dest="min_num_atoms", default=50, type=int,
        help="Minimum number of atoms")
    parser_supercell.add_argument(
        "--max_atoms", dest="max_num_atoms", default=300, type=int,
        help="Maximum number of atoms")

    parser_supercell.set_defaults(func=make_supercell)

    # -- defect_set ------------------------------------------------
    parser_defect_set = subparsers.add_parser(
        name="defect_set",
        description="Make defect_in.yaml file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['ds'])

    parser_defect_set.add_argument(
        "-o", "--oxi_states", nargs="+", type=str_int_to_int,
        help="Oxidation states in integers, e.g., Mg 2 O -2.")
    parser_defect_set.add_argument(
        "-d", "--dopants", nargs="+", type=str,
        help="Names of dopant element, e.g., Al Ga.")
    parser_defect_set.add_argument(
        "-k", "--kwargs", nargs="+", type=str,
        help="Keywords used to screen the target defects. "
             "Regular expression can be used.")

    parser_defect_set.set_defaults(func=make_defect_set)

    # -- defect_entries ------------------------------------------------
    parser_defect_entries = subparsers.add_parser(
        name="defect_entries",
        description="Make POSCAR files and defect entries in defect "
                    "directories",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['de'])

    parser_defect_entries.set_defaults(func=make_defect_entries)

    # -- append_interstitial ------------------------------------------------
    parser_append_interstitial = subparsers.add_parser(
        name="append_interstitial",
        description="Append interstitial information to supercell_info.yaml",
        parents=[si_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['ai'])

    parser_append_interstitial.add_argument(
        "-p", "--base_structure", required=True, type=Structure.from_file,
        help="POSCAR file used for generating AECCAR files, which must be the "
             "standardized primitive cell.")
    parser_append_interstitial.add_argument(
        "-c", "--frac_coords", required=True, nargs=3, type=float,
        help="Fractional coordinates in the standardized primitive cell added "
             "as an interstitial ")

    parser_append_interstitial.set_defaults(
        func=append_interstitial_to_supercell_info)

    # -- pop_interstitial ------------------------------------------------
    parser_pop_interstitial = subparsers.add_parser(
        name="pop",
        description="Pop interstitial site from supercell_info.json. ",
        parents=[si_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['pi'])

    parser_pop_interstitial.add_argument(
        "-i", "--index", required=True, type=int,
        help="Popped interstitial index. Note that the indices begin from 1.")

    parser_pop_interstitial.set_defaults(
        func=pop_interstitial_from_supercell_info)

    # -- calc_results ------------------------------------------------
    parser_calc_results = subparsers.add_parser(
        name="calc_results",
        description="Generate calc_results.json files. "
                    "Run after VASP calculations.",
        parents=[dirs_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cr'])
    parser_calc_results.set_defaults(func=make_calc_results)

    # -- efnv correction ------------------------------------------------
    parser_efnv = subparsers.add_parser(
        name="efnv",
        description="Generate extended FNV-correction related efnv_correction "
                    "files. ",
        parents=[dirs_parser, pcr_parser, unitcell_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['efnv'])

    parser_efnv.set_defaults(func=make_efnv_correction_from_vasp)

    # -- gkfo correction ------------------------------------------------
    parser_gkfo = subparsers.add_parser(
        name="gkfo",
        description="Generate GKFO correction files.",
        parents=[unitcell_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['gkfo'])

    parser_gkfo.add_argument(
        "-iefnv", "--initial_efnv_correction", required=True, type=loadfn,
        help="Path to the initial efnv correction.json file.")
    parser_gkfo.add_argument(
        "-icr", "--initial_calc_results", required=True, type=loadfn,
        help="Path to the initial calc_results.json file.")
    parser_gkfo.add_argument(
        "-fcr", "--final_calc_results", required=True, type=loadfn,
        help="Path to the final calc_results.json file.")
    parser_gkfo.add_argument(
        "-cd", "--charge_diff", required=True, type=int,
        help="Charge difference of final state from initial state.")

    parser_gkfo.set_defaults(func=make_gkfo_correction_from_vasp)

    # -- defect eigenvalues ------------------------------------------------
    parser_eig = subparsers.add_parser(
        name="eig",
        description="Generate plot of eigenvalues and occupations.",
        parents=[dirs_parser, pcr_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['eig'])

    parser_eig.set_defaults(func=make_defect_eigenvalues)

    # -- band edge characters ------------------------------------------------
    parser_ec = subparsers.add_parser(
        name="make_edge_characters",
        description="Make edge_characters.json files that are used for "
                    "analyzing the eigenvalues.",
        parents=[dirs_parser, pcr_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['mec'])

    parser_ec.set_defaults(func=make_edge_characters)

    # -- band edge states ------------------------------------------------
    parser_es = subparsers.add_parser(
        name="edge_states",
        description="Show edge state for each spin channel.",
        parents=[dirs_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['es'])

    parser_es.add_argument(
        "-p", "--perfect_edge_characters", required=True, type=loadfn,
        help="Path to the edge_characters.json for the perfect supercell.")

    parser_es.set_defaults(func=make_edge_states)

    # -- defect formation energy ----------------------------------------------
    parser_energy = subparsers.add_parser(
        name="defect_formation_energy",
        description="Show and plot defect formation energies.",
        parents=[dirs_parser, pcr_parser, unitcell_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['e'])

    parser_energy.add_argument(
        "-c", "--cpd_yaml", required=True, type=str,
        help="cpd.yaml file path.")
    parser_energy.add_argument(
        "-l", "--label", type=str, help="Label in cpd.pdf")
    parser_energy.add_argument(
        "-y", "--y_range", nargs=2, type=float,
        help="Energy range in y-axis")
    parser_energy.add_argument(
        "-s", "--skip_shallow", action="store_true",
        help="Set when the energies of shallow defects are omitted.")
    parser_energy.add_argument(
        "-p", "--print", action="store_true",
        help="Print the defect formation energy as a list.")
    parser_energy.add_argument(
        "--supercell_edge", action="store_true",
        help="Whether set the band edges to supercell ones instead of "
             "unitcell ones.")
    parser_energy.add_argument(
        "--no_label_line", dest="label_line", action="store_false",
        help="Set the labels not to locate on the lines.")
    parser_energy.add_argument(
        "--no_add_charges", dest="add_charges", action="store_false",
        help="Set the charges not to appear.")
    parser_energy.add_argument(
        "-wg", "--web_gui", action="store_true",
        help="Show web GUI.")
    parser_energy.add_argument(
        "--port", type=int, default=8050,
        help="Port for web GUI.")

    parser_energy.set_defaults(func=make_defect_formation_energy)

    # -- calc defect structure ------------------------------------------------
    parser_calc_def_str = subparsers.add_parser(
        name="calc_defect_structure_info",
        description="Calc defect structure.",
        parents=[dirs_parser, si_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['cds'])

    parser_calc_def_str.add_argument(
        "-dt", "--dist_tolerance", type=float, default=defaults.dist_tol,
        help="Tolerance in Angstrom for distance to judge if the "
             "atoms in initial and final structures are same ones.")
    parser_calc_def_str.add_argument(
        "--symprec", type=float,
        help="Tolerance for determining the point groups. If it's not set, "
             "the point groups written in defect_entry.json and "
             "calc_results.json are used.")
    parser_calc_def_str.set_defaults(func=calc_defect_structure_info)

    # ------------------------------------------------------------------------
    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import pytest
from pymatgen.io.vasp import Vasprun, Outcar

from pydefect.cli.vasp.make_calc_results import make_calc_results_from_vasp
from vise.tests.helpers.assertion import assert_msonable, assert_json_roundtrip


@pytest.fixture
def calc_results(vasp_files):
    vasprun = Vasprun(vasp_files / "MgO_conv_Va_O_0" / "vasprun.xml")
    outcar = Outcar(vasp_files / "MgO_conv_Va_O_0" / "OUTCAR")
    return make_calc_results_from_vasp(vasprun, outcar)


def test_calc_results_msonable(calc_results):
    assert_msonable(calc_results)


def test_calc_results_json_roundtrip(calc_results, tmpdir):
    assert_json_roundtrip(calc_results, tmpdir)


def test_calc_results_vbm_cbm(calc_results):
    assert calc_results.vbm == 2.4163
    assert calc_results.cbm == 10.2574




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import pytest
from pydefect.input_maker.add_interstitial import append_interstitial
from pydefect.input_maker.supercell_info import Interstitial
from pydefect.util.error_classes import NotPrimitiveError
from pymatgen.core import Structure, Lattice


def test_add_interstitial(cubic_supercell_info_wo_int):
    primitive = Structure(Lattice.rhombohedral(7.071068, 60),
                          species=["H", "He"],
                          coords=[[0.0]*3, [0.5]*3])
    new_supercell_info = append_interstitial(cubic_supercell_info_wo_int,
                                             primitive,
                                             [1/4, 1/4, 1/4])
    expected = Interstitial(frac_coords=[1/8, 1/8, 1/8],
                            wyckoff_letter="d",
                            site_symmetry="-43m")
    assert new_supercell_info.interstitials[0] == expected


def test_add_interstitial_not_primitive_error(cubic_supercell_info_wo_int):
    conventional_cell = Structure(Lattice.cubic(10),
                                  species=["H"] * 4 + ["He"] * 4,
                                  coords=[[0.0, 0.0, 0.0],
                                          [0.5, 0.5, 0.0],
                                          [0.5, 0.0, 0.5],
                                          [0.0, 0.5, 0.5],

                                          [0.0, 0.0, 0.5],
                                          [0.0, 0.5, 0.0],
                                          [0.5, 0.0, 0.0],
                                          [0.5, 0.5, 0.5],
                                        ])
    with pytest.raises(NotPrimitiveError):
        append_interstitial(cubic_supercell_info_wo_int, conventional_cell,
                            [1/4, 1/4, 1/4])




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, Tuple, Optional

from monty.json import MSONable
from monty.serialization import loadfn
from vise.util.enum import ExtendedEnum
from vise.util.mix_in import ToJsonFileMixIn


@dataclass
class BandEdgeEigenvalues(MSONable, ToJsonFileMixIn):
    # [spin, k-idx, band-idx] = energy, occupation
    energies_and_occupations: List[List[List[List[float]]]]
    kpt_coords: List[Tuple[float, float, float]]
    lowest_band_index: int


class EdgeState(MSONable, ExtendedEnum):
    donor_phs = "Donor PHS"
    acceptor_phs = "Acceptor PHS"
    in_gap_state = "In-gap state"
    no_in_gap = "No in-gap state"
    unknown = "Unknown"

    @property
    def is_shallow(self):
        return self in [self.acceptor_phs, self.donor_phs]


@dataclass
class BandEdgeStates(MSONable, ToJsonFileMixIn):
    states: List[EdgeState]  # by spin.

    @property
    def is_shallow(self):
        return any([s.is_shallow for s in self.states])

    def to_yaml(self, filename="band_edge_states.yaml"):
        d = [f"up: {self.states[0]}"]
        if len(self.states) == 2:
            d.append(f"down: {self.states[1]}")
        Path(filename).write_text("\n".join(d))

    @classmethod
    def from_yaml(cls, filename="band_edge_states.yaml"):
        d = loadfn(filename)
        states = [EdgeState(d.pop("up"))]
        if d:
            states.append(EdgeState(d.pop("down")))
        return cls(states)


class EdgeCharacters(Sequence, MSONable, ToJsonFileMixIn):
    def __init__(self, edge_characters: List["EdgeCharacter"]):
        self.edge_characters = edge_characters  # [by spin]

    def __getitem__(self, item):
        return self.edge_characters[item]

    def __len__(self):
        return len(self.edge_characters)

    def __str__(self):
        strings = []
        for spin, edge_char in zip(["up", "down"], self.edge_characters):
            strings.append(f"""{spin:>4}: {edge_char}""")
        return "\n".join(strings)


@dataclass
class EdgeCharacter(MSONable):
    """Code and its version dependent quantities. """
    hob_bottom_e: float  # Highest-occupied band
    lub_top_e: float  # Lowest-unoccupied band
    vbm: Optional[float]  # None for metals
    cbm: Optional[float]
    # {"Mn": [0.01, ..], "O": [0.03, 0.5]},
    # where lists contain s, p, d, (f) orbital components.
    vbm_orbitals: Dict[str, List[float]]
    cbm_orbitals: Dict[str, List[float]]
    vbm_participation_ratio: Optional[float] = None  # participation ratio
    cbm_participation_ratio: Optional[float] = None



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from typing import List

import numpy as np


def sanitize_matrix(matrix: List[int]) -> List[List[int]]:

    if len(matrix) == 9:
        return [matrix[:3], matrix[3:6], matrix[6:]]
    elif len(matrix) == 3:
        result = np.eye(3, dtype=int)
        for i in range(3):
            result[i, i] = matrix[i]
        return result.tolist()
    elif len(matrix) == 1:
        result = np.eye(3, dtype=int)
        for i in range(3):
            result[i, i] = matrix[0]
        return result.tolist()
    else:
        raise ValueError(f"Matrix element length {len(matrix)} is improper.")


def str_int_to_int(x):
    try:
        return int(x)
    except ValueError:
        return x



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.unitcell import Unitcell
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag
from pydefect.cli.vasp.main import parse_args
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection
from pydefect.defaults import defaults
from pydefect.input_maker.supercell_info import SupercellInfo
from pymatgen.core import Composition


def test_print(mocker):
    mock = mocker.patch("pydefect.cli.vasp.main.loadfn")
    parsed_args = parse_args(["p", "-f", "a.json"])
    expected = Namespace(obj=mock.return_value, func=parsed_args.func)
    assert parsed_args == expected


def test_unitcell(mocker):
    mock = mocker.patch("pydefect.cli.vasp.main.Vasprun")
    mock_outcar = mocker.patch("pydefect.cli.vasp.main.Outcar")

    parsed_args = parse_args(["u",
                              "-vb", "vasprun.xml",
                              "-ob", "OUTCAR-1",
                              "-odc", "OUTCAR-2",
                              "-odi", "OUTCAR-3"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        vasprun_band=mock.return_value,
        outcar_band=mock_outcar.return_value,
        outcar_dielectric_clamped=mock_outcar.return_value,
        outcar_dielectric_ionic=mock_outcar.return_value,
        func=parsed_args.func,
    )
    assert parsed_args == expected
    mock.assert_called_once_with("vasprun.xml")
    mock_outcar.assert_any_call("OUTCAR-1")
    mock_outcar.assert_any_call("OUTCAR-2")
    mock_outcar.assert_called_with("OUTCAR-3")


def test_make_poscars_wo_options():
    parsed_args = parse_args(["mp",
                              "-e", "Mg", "O"])
    expected = Namespace(
        elements=["Mg", "O"],
        e_above_hull=defaults.e_above_hull,
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_poscars_w_options():
    parsed_args = parse_args(["mp",
                              "-e", "Mg", "O",
                              "--e_above_hull", "0.1"])
    expected = Namespace(
        elements=["Mg", "O"],
        e_above_hull=0.1,
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_cpd_options():
    parsed_args = parse_args(["mcpd",
                              "-d", "Mg", "O",
                              "-t", "MgO"])
    expected = Namespace(
        dirs=[Path("Mg"), Path("O")],
        target=Composition("MgO"),
        elements=None,
        atom_energy_yaml=None,
        yaml="cpd.yaml",
        update=False,
        vertices=None,
        func=parsed_args.func)
    assert parsed_args == expected


def test_plot_cpd_options():
    parsed_args = parse_args(["pcpd",
                              "-y", "a.yaml"])
    expected = Namespace(
        yaml="a.yaml",
        func=parsed_args.func)
    assert parsed_args == expected


def test_make_supercell_wo_options(mocker):
    mock = mocker.patch("pydefect.cli.vasp.main.IStructure")
    parsed_args = parse_args(["s"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        unitcell=mock.from_file.return_value,
        matrix=None,
        min_num_atoms=50,
        max_num_atoms=300,
        func=parsed_args.func,
    )
    assert parsed_args == expected
    mock.from_file.assert_called_once_with("POSCAR")


def test_make_supercell_w_options(mocker):
    mock = mocker.patch("pydefect.cli.vasp.main.IStructure")
    parsed_args = parse_args(["s",
                              "-p", "POSCAR-tmp",
                              "--matrix", "1", "2", "3",
                              "--min_atoms", "1000",
                              "--max_atoms", "2000"])
    # func is a pointer so need to point the same address.
    expected = Namespace(
        unitcell=mock.from_file.return_value,
        matrix=[1, 2, 3],
        min_num_atoms=1000,
        max_num_atoms=2000,
        func=parsed_args.func,
    )
    assert parsed_args == expected
    mock.from_file.assert_called_once_with("POSCAR-tmp")


def test_defect_set_wo_options():
    parsed_args = parse_args(["ds"])
    expected = Namespace(
        oxi_states=None,
        dopants=None,
        kwargs=None,
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_defect_set_w_options():
    parsed_args = parse_args(["ds",
                              "-o", "He", "1",
                              "-d", "Li",
                              "-k", "Li_H1", "Va_H1_0"])
    expected = Namespace(
        oxi_states=["He", 1],
        dopants=["Li"],
        kwargs=["Li_H1", "Va_H1_0"],
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_defect_entries():
    parsed_args = parse_args(["de"])
    expected = Namespace(
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_append_interstitial(mocker):

    mock_loadfn = mocker.patch("pydefect.cli.vasp.main.loadfn")
    mock_supercell_info = mocker.Mock(spec=SupercellInfo, autospec=True)
    mock_loadfn.return_value = mock_supercell_info
    mock_structure = mocker.patch("pydefect.cli.vasp.main.Structure")
    parsed_args = parse_args(["ai",
                              "-s", "supercell_info.json",
                              "-p", "POSCAR",
                              "-c", "0.1", "0.2", "0.3"])
    expected = Namespace(
        supercell_info=mock_supercell_info,
        base_structure=mock_structure.from_file.return_value,
        frac_coords=[0.1, 0.2, 0.3],
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_pop_interstitial(mocker):

    mock_loadfn = mocker.patch("pydefect.cli.vasp.main.loadfn")
    mock_supercell_info = mocker.Mock(spec=SupercellInfo, autospec=True)
    mock_loadfn.return_value = mock_supercell_info
    parsed_args = parse_args(["pi",
                              "-s", "supercell_info.json",
                              "-i", "1000"])
    expected = Namespace(
        supercell_info=mock_supercell_info,
        index=1000,
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_calc_results():
    parsed_args = parse_args(["cr", "-d", "Va_O1_0", "Va_O1_1"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        func=parsed_args.func,
    )
    assert parsed_args == expected


def test_efnv_correction(mocker):
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_unitcell = mocker.Mock(spec=Unitcell, autospec=True)

    def side_effect(filename):
        if filename == "perfect/calc_results.json":
            return mock_calc_results
        elif filename == "unitcell.json":
            return mock_unitcell
        else:
            raise ValueError

    mocker.patch("pydefect.cli.vasp.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args(["efnv",
                              "-d", "Va_O1_0", "Va_O1_1",
                              "-pcr", "perfect/calc_results.json",
                              "-u", "unitcell.json"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        perfect_calc_results=mock_calc_results,
        unitcell=mock_unitcell,
        func=parsed_args.func)
    assert parsed_args == expected


def test_gkfo_correction(mocker):
    mock_i_correction = mocker.Mock(spec=ExtendedFnvCorrection, autospec=True)
    mock_i_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_f_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_unitcell = mocker.Mock(spec=Unitcell, autospec=True)

    def side_effect(filename):
        if filename == "a/correction.json":
            return mock_i_correction
        elif filename == "a/calc_results.json":
            return mock_i_calc_results
        elif filename == "a/absorption/calc_results.json":
            return mock_f_calc_results
        elif filename == "unitcell.json":
            return mock_unitcell
        else:
            raise ValueError

    mocker.patch("pydefect.cli.vasp.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args(["gkfo",
                              "-iefnv", "a/correction.json",
                              "-cd", "1",
                              "-icr", "a/calc_results.json",
                              "-fcr", "a/absorption/calc_results.json",
                              "-u", "unitcell.json"])
    expected = Namespace(
        initial_efnv_correction=mock_i_correction,
        initial_calc_results=mock_i_calc_results,
        final_calc_results=mock_f_calc_results,
        charge_diff=1,
        unitcell=mock_unitcell,
        func=parsed_args.func)
    assert parsed_args == expected


def test_defect_eigenvalues(mocker):
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)

    def side_effect(filename):
        if filename == "perfect/calc_results.json":
            return mock_calc_results
        else:
            raise ValueError

    mocker.patch("pydefect.cli.vasp.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args(["eig",
                              "-d", "Va_O1_0", "Va_O1_1",
                              "-pcr", "perfect/calc_results.json"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        perfect_calc_results=mock_calc_results,
        func=parsed_args.func)
    assert parsed_args == expected


def test_band_edge_characters(mocker):
    mock = mocker.patch("pydefect.cli.vasp.main.loadfn")
    parsed_args = parse_args(["mec",
                              "-d", "Va_O1_0", "Va_O1_1",
                              "-pcr", "perfect/calc_results.json"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        perfect_calc_results=mock.return_value,
        func=parsed_args.func)
    assert parsed_args == expected
    mock.assert_any_call("perfect/calc_results.json")


def test_band_edge_states(mocker):
    mock = mocker.patch("pydefect.cli.vasp.main.loadfn")
    parsed_args = parse_args(["es",
                              "-d", "Va_O1_0", "Va_O1_1",
                              "-p", "perfect/edge_characters.json"])
    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        perfect_edge_characters=mock.return_value,
        func=parsed_args.func)
    assert parsed_args == expected
    mock.assert_any_call("perfect/edge_characters.json")


def test_defect_formation_energy(mocker):
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_unitcell = mocker.Mock(spec=Unitcell, autospec=True)
    mock_chem_pot_diag = mocker.Mock(spec=ChemPotDiag, autospec=True)

    def side_effect(filename):
        if filename == "perfect/calc_results.json":
            return mock_calc_results
        elif filename == "unitcell.json":
            return mock_unitcell
        else:
            raise ValueError

    mocker.patch("pydefect.cli.vasp.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args(["e",
                              "-d", "Va_O1_0", "Va_O1_1",
                              "-pcr", "perfect/calc_results.json",
                              "-u", "unitcell.json",
                              "-c", "cpd.yaml",
                              "-l", "A",
                              "-y", "-5", "5",
                              "-s",
                              "-p"
                              ])

    expected = Namespace(
        dirs=[Path("Va_O1_0"), Path("Va_O1_1")],
        perfect_calc_results=mock_calc_results,
        unitcell=mock_unitcell,
        cpd_yaml="cpd.yaml",
        label="A",
        y_range=[-5, 5],
        skip_shallow=True,
        print=True,
        supercell_edge=False,
        label_line=True,
        add_charges=True,
        web_gui=False,
        port=8050,
        func=parsed_args.func)
    assert parsed_args == expected


def test_defect_structure(mocker):
    mock_supercell_info = mocker.Mock(spec=SupercellInfo, autospec=True)

    def side_effect(filename):
        if filename == "supercell_info.json":
            return mock_supercell_info
        else:
            raise ValueError

    mock = mocker.patch("pydefect.cli.vasp.main.loadfn", side_effect=side_effect)
    parsed_args = parse_args(["cds",
                              "-s", "supercell_info.json",
                              "-d", "Va_O1_0",
                              "-dt", "1.0",
                              "--symprec", "2.0"])
    expected = Namespace(
        supercell_info=mock_supercell_info,
        dirs=[Path("Va_O1_0")],
        dist_tolerance=1.0,
        symprec=2.0,
        func=parsed_args.func)
    assert parsed_args == expected



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest

from pydefect.cli.main_tools import sanitize_matrix, str_int_to_int


def test_sanitize_matrix_9_input_values():
    actual = sanitize_matrix(list(range(9)))
    expected = [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    assert actual == expected


def test_sanitize_matrix_3_input_values():
    actual = sanitize_matrix(list(range(1, 4)))
    expected = [[1, 0, 0], [0, 2, 0], [0, 0, 3]]
    assert actual == expected


def test_sanitize_matrix_1_input_value():
    actual = sanitize_matrix([4])
    expected = [[4, 0, 0], [0, 4, 0], [0, 0, 4]]
    assert actual == expected


def test_sanitize_matrix_raise_error_for_improper_matrix_element_length():
    with pytest.raises(ValueError):
         sanitize_matrix([1, 4])


def test_str_int_to_int():
    assert str_int_to_int("a") == "a"
    assert str_int_to_int("1") == 1
    assert str_int_to_int("1.1") == "1.1"


"""
TODO
- return matrix

DONE
"""


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import argparse
import sys

from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag, CpdPlotInfo
from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiagPlotly3DMplPlotter


def parse_args(args):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-c", "--chem_pot_diag", type=str)
    parser.add_argument("--port", type=int)
    return parser.parse_args(args)


if __name__ == "__main__":
    args = parse_args(sys.argv[1:])
    cpd = ChemPotDiag.from_yaml(args.chem_pot_diag)
    print(cpd.target_vertices)
    print(cpd.vertex_coords)
    cpd_plot_info = CpdPlotInfo(cpd)
    print(cpd_plot_info.comp_vertices)
    plotter = ChemPotDiagPlotly3DMplPlotter(cpd_plot_info)
    fig = plotter.figure
    fig.show()
#    ctc.register_crystal_toolkit(app=app, layout=layout, cache=None)
#    app.run_server(debug=True, port=args.port)


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import shutil

import fire
from monty.serialization import loadfn
from pydefect.analyzer.defect_structure_symmetrizer import \
    symmetrize_defect_structure
from pydefect.defaults import defaults
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import Structure
from vise.util.structure_symmetrizer import StructureSymmetrizer


def make_refined_structure():
    defect_entry: DefectEntry = loadfn("defect_entry.json")
    structure = Structure.from_file(defaults.contcar)
    symmetrizer = StructureSymmetrizer(structure,
                                       defaults.symmetry_length_tolerance,
                                       defaults.symmetry_angle_tolerance)
    if symmetrizer.point_group == "1":
        return

    shutil.move(defaults.contcar, str(defaults.contcar) + ".sym_1")
    shutil.move(defaults.outcar, str(defaults.outcar) + ".sym_1")
    shutil.move(defaults.vasprun, str(defaults.vasprun) + ".sym_1")

    refined_structure = symmetrize_defect_structure(
        symmetrizer,
        defect_entry.anchor_atom_index,
        defect_entry.anchor_atom_coords)

    refined_structure.to(fmt="POSCAR", filename="POSCAR")


if __name__ == '__main__':
    fire.Fire(make_refined_structure)
